/*
 * $Id$
 *
 * Author: Daniel Silverstone <dsilvers@digital-scurf.org>
 *
 * Purpose: Lua ISAPI runner - Runs *.lua and outputs the write'd output,
 *          If extension is not .lua, it assumes that it is an HTML
 *          file with lua code embedded in (unprocessed) <<LUA<< >>LUA>>
 */

#include <sys/types.h>
#include <unistd.h>

/* Our LUA functions*/
#include "isapi_lua.h"

/* Import the ISAPI constants and type definitions */
#include "httpext.h"

/* Our cache code */
#include "cache.h"
#include "settings.h"

/* standard libs */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

BOOL WINAPI 
GetExtensionVersion( HSE_VERSION_INFO *pVer ) 
{
  pVer->dwExtensionVersion = HSE_VERSION_MAJOR;
  strncpy(pVer->lpszExtensionDesc, "Digital-scurf Content generator", 
	  HSE_MAX_EXT_DLL_NAME_LEN);
  return TRUE;
}

int lua_getservervariable(lua_State *L)
{
  LPEXTENSION_CONTROL_BLOCK ecb = (LPEXTENSION_CONTROL_BLOCK)(lua_touserdata(L,-1));
  char buffer[2048];
  int length;
  length = 2048;
  DebugOut("Fetching %s...\n",lua_tostring(L,1));
  if (ecb->GetServerVariable( ecb->ConnID, lua_tostring(L,1), buffer, &length ))
    {
      lua_settop(L,0);
      length--;
      lua_pushlstring(L,buffer,length);
      DebugOut("Returning '%s' which is %d bytes\n",buffer,length);
    }
  else
    {
      lua_settop(L,0);
      lua_pushstring(L,"");
    }
  return 1; /* the value */
}

int lua_translatepath(lua_State *L)
{
  LPEXTENSION_CONTROL_BLOCK ecb = (LPEXTENSION_CONTROL_BLOCK)(lua_touserdata(L,-1));
  char buffer[2048];
  int length = 2048;
  strcpy(buffer,lua_tostring(L,1));
  DebugOut("Translating %s...\n",lua_tostring(L,1));
  ecb->ServerSupportFunction( ecb->ConnID, HSE_REQ_MAP_URL_TO_PATH, buffer, &length, 0 );
  lua_pushstring(L,buffer);
  DebugOut("Returning '%s'\n",buffer);
  return 1;
}

void slurp_and_run(lua_State *L, const char* filename, LPEXTENSION_CONTROL_BLOCK ecb)
{
  int cacheentry = -1;
  const unsigned char* filedata;
  int length;
  char buffer[1500];
  unsigned char *copybuffer;
  int own = 0;
  length = 1024;
  buffer[0] = '/'; buffer[1] = 0;
  ecb->ServerSupportFunction( ecb->ConnID, HSE_REQ_MAP_URL_TO_PATH, buffer, &length, 0 );
  lua_pushstring(L,buffer);
  lua_setglobal(L,"_DOCROOT");
  skip = strlen(buffer)-1;
  strcat(buffer,"../lua/module.lua");
  lua_pushstring(L,filename);
  lua_setglobal(L,"_INPUT_FILE");
  LockCache();
  cacheentry = GetFile(buffer,1); /* The globals file *is* "important" */
  filedata = GetData(cacheentry,1);
  length = GetSize(cacheentry,1);
  if (length == -1)
    {
      copybuffer = (unsigned char *)"error(\"Unable to find global module lua file\")";
      length = strlen(filedata);
      own = 1;
    }
  else
    {
      copybuffer = (unsigned char *)(malloc(length));
      memcpy(copybuffer,filedata,length);
    }
  UnlockCache();
  /* Now that we're unlocked, we can run the code... */
  lua_dobuffer(L, copybuffer, length, "Global file");
  /* Extract the contents of the various bits */
  lua_settop(L,0);
  lua_getglobal(L,"HEADER_CONTENTS");
  lua_getglobal(L,"RESPONSE");
  lua_getglobal(L,"_OUTPUT");
  /* and do the right thing (TM) */
  cacheentry = lua_strlen(L,1);

  ecb->ServerSupportFunction( ecb->ConnID, 
			      HSE_REQ_SEND_RESPONSE_HEADER, 
			      (char*)(lua_tostring(L,2)),  
			      &cacheentry, 
			      (DWORD *)(lua_tostring(L,1)) );
  /* Body... */
  cacheentry = lua_strlen(L,3);
  ecb->WriteClient( ecb->ConnID, (char*)(lua_tostring(L,3)), &cacheentry,0 );
  if (!own)
    {
      free(copybuffer);
    }
}

DWORD WINAPI
HttpExtensionProc( LPEXTENSION_CONTROL_BLOCK ecb )
{
  char variablebuffer[2048];
  int varlen = 2048;


  /* Initialise a lua state */
  lua_State *control;

  DebugOut("Running a connection\n");

  control = ISAPI_Create_Lua_State( ecb );
  register_cache(control);

  lua_pushusertag(control, ecb, 0);
  lua_pushcclosure(control, lua_getservervariable, 1);
  lua_setglobal(control,"GetServerVariable");

  lua_pushusertag(control, ecb, 0);
  lua_pushcclosure(control, lua_translatepath, 1);
  lua_setglobal(control,"TranslateServerPath");

  /* Now we need to work out what script to run */
  ecb->GetServerVariable( ecb->ConnID, "PATH_TRANSLATED", variablebuffer, &varlen );
  
  /* For now, simply lua_dofile() that file */

  if (varlen)
    {
      slurp_and_run(control, variablebuffer, ecb); 
    }
  else
    {
      lua_dostring(control, "Send([[No data]])");
    }
  lua_close(control);
  DebugOut("Completed\n");
  /* return, indicating success */
  return HSE_STATUS_SUCCESS;   
}


int DllMain(void *ignore1, int notification, void *ignore2)
{
  if (notification == DLL_PROCESS_ATTACH)
    {
      InitCaches();
    }
  if (notification == DLL_PROCESS_DETACH)
    {
      fprintf(stderr,"Clear cache...\n");fflush(stderr);
      ClearCaches();
    }
  return 1;
}
